/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2002 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: xcolorsel.cc,v 1.1 2002/09/09 18:31:03 itz Exp $ ";

#include "xcolorsel.hh"
#include "my_color_selection.hh"
#include "cursor_image.xbm"
#include "cursor_mask.xbm"

#include <gtk--/main.h>
#include <gdk--/color.h>
#include <gdk--/pixmap.h>
#include <gdk--/window.h>
#include <gdk--/colormap.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdktypes.h>
#include <X11/X.h>
#include <math.h>

// this algorithm is more or less copied straight from the original
// Xcolorsel.  While the gdk function gdk_window_xid_at_coords() seems
// to do something similar, it is completely undocumented, labelled
// "mostly for internal use" (?) and also, it seems, buggy: it can
// return input-only windows.

// I guess this means Windows users are SOL, but it is not by design
// this once.  I bet Qt does better.

static Window 
x_window_from_coords(gint xabs, gint yabs, Display* p_d, Window root)
{
  gint new_x = xabs;
  gint new_y = yabs;
  Window new_window = root;
  Window last_window = root;
  Window last_real_window = root;
  Window child = None;

  XGrabServer(p_d);
  XTranslateCoordinates(p_d, last_window, new_window, new_x, new_y,
                        &new_x, &new_y, &child);
  while (child != None) {
    XWindowAttributes atb;
    XGetWindowAttributes(p_d, child, &atb);
    if (atb.c_class != InputOnly) {
      last_real_window = child;
    } /*if*/
    last_window = new_window;
    new_window = child;
    XTranslateCoordinates(p_d, last_window, new_window, new_x, new_y,
                          &new_x, &new_y, &child);
  } /*while*/
  XUngrabServer(p_d);
  return last_real_window;
}

void
xcolorsel_window::destroy_handler()
{
  Gtk::Main::quit();
}

xcolorsel_window::xcolorsel_window()
  : Gtk::Window(GTK_WINDOW_TOPLEVEL), grabbing(0)
{
  set_title ("xcolorsel-gtk");
  set_border_width(10);
  destroy.connect(SigC::slot(destroy_handler));
  p_color_selection = manage(new my_color_selection);
  add(*p_color_selection);
  show_all();
  p_color_selection->set_opacity(0);
}

void
xcolorsel_window::start_grab_loop(void)
{
  Gdk_Colormap cm = Gdk_Colormap::get_system();
  Gdk_Color black = cm.black();
  Gdk_Color white = cm.white();
  Gdk_Pixmap cursor_image(get_window(), cursor_image_bits, cursor_image_width,
                          cursor_image_height, 1, black, white);
  Gdk_Pixmap cursor_mask(get_window(), cursor_mask_bits, cursor_mask_width,
                         cursor_mask_height, 1, white, black);
  Gdk_Cursor cursor(cursor_image, cursor_mask, black, white, 10, 11);
  if (GrabSuccess !=
      gdk_pointer_grab(get_window(), 0,
                         (GdkEventMask)(GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK),
                         0, cursor, GDK_CURRENT_TIME)) {
    cerr << "Failed to grab cursor" << endl;
    return;
  } /*if*/
  //  gdk_set_show_events(1);
  grabbing = 1;
}

gint
xcolorsel_window::button_press_event_impl(GdkEventButton* p0)
{
  if (grabbing && (p0->type == GDK_BUTTON_PRESS)) {
    gint xabs = (gint)trunc(p0->x_root);
    gint yabs = (gint)trunc(p0->y_root);
    ::Window xid = x_window_from_coords(xabs, yabs, gdk_display, gdk_root_window);
    GdkWindow* wp = gdk_window_lookup(xid);
    if (wp == 0) {
      wp = gdk_window_foreign_new
        (x_window_from_coords(xabs, yabs, gdk_display, gdk_root_window));
    } /*if*/
    Gdk_Window w(wp);
    gint xorigin = 0;
    gint yorigin = 0;
    w.get_origin(xorigin, yorigin);
    gint xrel = xabs - xorigin;
    gint yrel = yabs - yorigin;
    Gdk_Image pixel_image;
    pixel_image.get(w, xrel, yrel, 1, 1);
    GdkColorContext* cc = gdk_color_context_new(pixel_image->visual, w.get_colormap());
    GdkColor gdkc;
    gdkc.pixel = pixel_image.get_pixel(0, 0);
    gdk_color_context_query_color(cc, &gdkc);
    gdk_color_context_free(cc);
    gdouble grabbed_color[4];
    grabbed_color[0] = (gdouble)gdkc.red/(gdouble)65535;
    grabbed_color[1] = (gdouble)gdkc.green/(gdouble)65535;
    grabbed_color[2] = (gdouble)gdkc.blue/(gdouble)65535;
    grabbed_color[3] = 0.0;
    p_color_selection->set_color_and_notify(grabbed_color);
    //gdk_window_unref(wp);
    // gdk_set_show_events(0);
    grabbing = 0;
    gdk_pointer_ungrab(GDK_CURRENT_TIME);
    return 1;
  } else {
    return inherited::button_press_event_impl(p0);
  } /*if*/
}
